using System;
using System.Reflection.Emit;
using Collections = System.Collections.Generic;
using Reflect = System.Reflection;
using Emit = System.Reflection.Emit;
using IO = System.IO;
using System.Text.RegularExpressions;

public sealed class CodeGen {
    Emit.ILGenerator il = null;
    Collections.Dictionary<string, Emit.LocalBuilder> symbolTable;

    public CodeGen(Stmt stmt, string moduleName) {
        if (IO.Path.GetFileName(moduleName) != moduleName) {
            throw new System.Exception("can only output into current directory!");
        }

        Reflect.AssemblyName name = new Reflect.AssemblyName(IO.Path.GetFileNameWithoutExtension(moduleName));
        Emit.AssemblyBuilder asmb = System.AppDomain.CurrentDomain.DefineDynamicAssembly(name, Emit.AssemblyBuilderAccess.Save);
        Emit.ModuleBuilder modb = asmb.DefineDynamicModule(moduleName);
        Emit.TypeBuilder typeBuilder = modb.DefineType("Foo");

        Emit.MethodBuilder methb = typeBuilder.DefineMethod("Main", Reflect.MethodAttributes.Static, typeof(void), System.Type.EmptyTypes);

        // CodeGenerator
        this.il = methb.GetILGenerator();
        this.symbolTable = new Collections.Dictionary<string, Emit.LocalBuilder>();

        // Go Compile!
        this.GenStmt(stmt);

        il.Emit(Emit.OpCodes.Ret);
        typeBuilder.CreateType();
        modb.CreateGlobalFunctions();
        asmb.SetEntryPoint(methb);
        asmb.Save(moduleName);
        this.symbolTable = null;
        this.il = null;
    }


    private void GenStmt(Stmt stmt) {
        if (stmt is Sequence) {
            Sequence seq = (Sequence)stmt;
            this.GenStmt(seq.First);
            this.GenStmt(seq.Second);
        }

        else if (stmt is DeclareVar) {
            // declare a local
            DeclareVar declare = (DeclareVar)stmt;
            this.symbolTable[declare.Ident] = this.il.DeclareLocal(this.TypeOfExpr(declare.Expr));

            // set the initial value
            Assign assign = new Assign();
            assign.Ident = declare.Ident;
            assign.Expr = declare.Expr;
            this.GenStmt(assign);
        }
        else if (stmt is Assign) {
            Assign assign = (Assign)stmt;
            this.GenExpr(assign.Expr, this.TypeOfExpr(assign.Expr));
            this.Store(assign.Ident, this.TypeOfExpr(assign.Expr));
        }
        else if (stmt is Print) {
            // the "print" statement is an alias for System.Console.WriteLine. 
            // it uses the string case
            this.GenExpr(((Print)stmt).Expr, typeof(string));
            this.il.Emit(Emit.OpCodes.Call, typeof(System.Console).GetMethod("WriteLine", new System.Type[] { typeof(object) }));
        }
        else if (stmt is ReadInt) {
            this.il.Emit(Emit.OpCodes.Call, typeof(System.Console).GetMethod("ReadLine", System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static, null, new System.Type[] { }, null));
            this.il.Emit(Emit.OpCodes.Call, typeof(int).GetMethod("Parse", System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Static, null, new System.Type[] { typeof(string) }, null));
            this.Store(((ReadInt)stmt).Ident, typeof(int));
        }
        else if (stmt is ForLoop) {
            // example: 
            // for x = 0 to 100 do
            //   print "hello";
            // end;

            // x = 0
            ForLoop forLoop = (ForLoop)stmt;
            Assign assign = new Assign();
            assign.Ident = forLoop.Ident;
            assign.Expr = forLoop.From;
            this.GenStmt(assign);
            // jump to the test
            Emit.Label test = this.il.DefineLabel();
            this.il.Emit(Emit.OpCodes.Br, test);

            // statements in the body of the for loop
            Emit.Label body = this.il.DefineLabel();
            this.il.MarkLabel(body);
            this.GenStmt(forLoop.Body);

            // to (increment the value of x)
            this.il.Emit(Emit.OpCodes.Ldloc, this.symbolTable[forLoop.Ident]);
            this.il.Emit(Emit.OpCodes.Ldc_I4, 1);
            this.il.Emit(Emit.OpCodes.Add);
            this.Store(forLoop.Ident, typeof(int));

            // **test** does x equal 100? (do the test)
            this.il.MarkLabel(test);
            this.il.Emit(Emit.OpCodes.Ldloc, this.symbolTable[forLoop.Ident]);
            this.GenExpr(forLoop.To, typeof(int));
            this.il.Emit(Emit.OpCodes.Blt, body);
        }
        else if (stmt is Foreach) {
            Foreach each = (Foreach)stmt;
            Label body = il.DefineLabel();
            Label test = il.DefineLabel();

            il.Emit(OpCodes.Br, test);

            il.MarkLabel(body);
            GenExpr(each.IEnumerable, typeof(System.Collections.IEnumerator));
            il.Emit(OpCodes.Callvirt, typeof(System.Collections.IEnumerator).GetMethod("get_Current"));
            il.Emit(OpCodes.Castclass, typeof(System.Text.RegularExpressions.Match));
            il.Emit(OpCodes.Callvirt, typeof(Capture).GetMethod("get_Value"));
            symbolTable[each.Ident] = il.DeclareLocal(typeof(object));
            Store(each.Ident, typeof(object));

            GenStmt(each.Body);

            il.MarkLabel(test);
            GenExpr(each.IEnumerable, typeof(System.Collections.IEnumerator));
            il.Emit(OpCodes.Callvirt, typeof(System.Collections.IEnumerator).GetMethod("MoveNext"));
            il.Emit(OpCodes.Brtrue, body);

        }
        else if (stmt is If) {
            If ifstmt = (If)stmt;
            Label elseLabel = il.DefineLabel();

            GenExpr(ifstmt.Condition, typeof(bool));
            il.Emit(OpCodes.Ldc_I4_0);
            il.Emit(OpCodes.Ceq);
            il.Emit(OpCodes.Brtrue, elseLabel);

            GenStmt(ifstmt.Body);
            il.MarkLabel(elseLabel);
        }
        else if (stmt is Append) {
            Append append = (Append)stmt;
            GenExpr(append.Buider, typeof(System.Text.StringBuilder));
            GenExpr(append.ToAppend, typeof(string));
            il.Emit(OpCodes.Callvirt, typeof(System.Text.StringBuilder).GetMethod("Append",
                new[] { typeof(string) }));
            il.Emit(OpCodes.Pop);
        }
        else {
            throw new System.Exception("don't know how to gen a " + stmt.GetType().Name);
        }
    }

    private void Store(string name, System.Type type) {
        if (this.symbolTable.ContainsKey(name)) {
            Emit.LocalBuilder locb = this.symbolTable[name];

            if (locb.LocalType == type) {
                this.il.Emit(Emit.OpCodes.Stloc, this.symbolTable[name]);
            }
            else {
                throw new System.Exception("'" + name + "' is of type " + locb.LocalType.Name + " but attempted to store value of type " + type.Name);
            }
        }
        else {
            throw new System.Exception("undeclared variable '" + name + "'");
        }
    }

    private void GenExpr(Expr expr, System.Type expectedType) {
        System.Type deliveredType = null;

        if (expr is StringLiteral) {
            deliveredType = typeof(string);
            this.il.Emit(Emit.OpCodes.Ldstr, ((StringLiteral)expr).Value);
        }
        else if (expr is IntLiteral) {
            deliveredType = typeof(int);
            this.il.Emit(Emit.OpCodes.Ldc_I4, ((IntLiteral)expr).Value);
        }
        else if (expr is Variable) {
            string ident = ((Variable)expr).Ident;
            deliveredType = this.TypeOfExpr(expr);

            if (!this.symbolTable.ContainsKey(ident)) {
                throw new System.Exception("undeclared variable '" + ident + "'");
            }

            this.il.Emit(Emit.OpCodes.Ldloc, this.symbolTable[ident]);
        }
        else if (expr is Match) {
            Match m = (Match)expr;
            GenExpr(m.Pattern, typeof(string));
            il.Emit(OpCodes.Newobj, typeof(Regex).GetConstructor(new[] { typeof(string) }));
            GenExpr(m.Input, typeof(string));
            il.Emit(OpCodes.Callvirt, typeof(Regex).GetMethod("Matches", new[] { typeof(string) }));
            il.Emit(OpCodes.Callvirt, typeof(MatchCollection).GetMethod("GetEnumerator"));
            deliveredType = typeof(System.Collections.IEnumerator);
        }
        else if (expr is StrLen) {
            StrLen len = (StrLen)expr;
            GenExpr(len.Input, typeof(string));
            il.Emit(OpCodes.Callvirt, typeof(string).GetMethod("get_Length"));
            deliveredType = typeof(int);
        }
        else if (expr is BinExpr) {
            BinExpr bin = (BinExpr)expr;
            if (bin.Op == BinOp.Eq) {
                GenExpr(bin.Left, typeof(int));
                GenExpr(bin.Right, typeof(int));
                il.Emit(OpCodes.Ceq);
                deliveredType = typeof(bool);
            }
            else {
                throw new Exception("binexpr unknow op");
            }
        }
        else if (expr is Builder) {
            Builder builder = (Builder)expr;
            il.Emit(OpCodes.Newobj, typeof(System.Text.StringBuilder)
                .GetConstructor(new Type[] { }));
            deliveredType = typeof (System.Text.StringBuilder);
        }
        else {
            throw new System.Exception("don't know how to generate " + expr.GetType().Name);
        }

        if (deliveredType != expectedType) {
            if (deliveredType == typeof(int) &&
                expectedType == typeof(string)) {
                this.il.Emit(Emit.OpCodes.Box, typeof(int));
                this.il.Emit(Emit.OpCodes.Callvirt, typeof(object).GetMethod("ToString"));
            }
            else if (expectedType == typeof(string)) {
                this.il.Emit(OpCodes.Callvirt, typeof(object).GetMethod("ToString"));
            }
            else {
                throw new System.Exception("can't coerce a " + deliveredType.Name + " to a " + expectedType.Name);
            }
        }

    }

    private System.Type TypeOfExpr(Expr expr) {
        if (expr is StringLiteral) {
            return typeof(string);
        }
        else if (expr is IntLiteral) {
            return typeof(int);
        }
        else if (expr is Match) {
            return typeof(System.Collections.IEnumerator);
        }
        else if (expr is StrLen) {
            return typeof(int);
        }
        else if (expr is BinExpr && ((BinExpr)expr).Op == BinOp.Eq) {
            return typeof(bool);
        }
        else if (expr is Builder) {
            return typeof(System.Text.StringBuilder);
        }
        else if (expr is Variable) {
            Variable var = (Variable)expr;
            if (this.symbolTable.ContainsKey(var.Ident)) {
                Emit.LocalBuilder locb = this.symbolTable[var.Ident];
                return locb.LocalType;
            }
            else {
                throw new System.Exception("undeclared variable '" + var.Ident + "'");
            }
        }
        else {
            throw new System.Exception("don't know how to calculate the type of " + expr.GetType().Name);
        }
    }
}
