using System;
using System.Collections.Generic;
using System.Text;
using Collections = System.Collections.Generic;

public sealed class Parser {
    private readonly Stmt result;
    private readonly IList<object> tokens;
    private int index;

    public Parser(IList<object> tokens) {
        this.tokens = tokens;
        index = 0;
        result = ParseStmt();

        if (index != this.tokens.Count)
            throw new Exception("expected EOF");
    }

    public Stmt Result {
        get { return result; }
    }

    private Stmt ParseStmt() {
        Stmt result;

        if (index == tokens.Count) {
            throw new Exception("expected statement, got EOF");
        }

        // <stmt> := print <expr> 

        // <expr> := <string>
        // | <int>
        // | <arith_expr>
        // | <ident>
        if (tokens[index].Equals("print")) {
            index++;
            var print = new Print();
            print.Expr = ParseExpr();
            result = print;
        }
        else if (tokens[index].Equals("append"))
        {
            index++;
            var append = new Append();
            if (index < tokens.Count) {
                append.Buider = ParseExpr();
            }
            else {
                throw new Exception("expected Buider after 'append'");
            }
            if (index < tokens.Count) {
                append.ToAppend = ParseExpr();
            }
            else {
                throw new Exception("expected ToAppend after 'append'");
            }
            result = append;
        }
        else if (tokens[index].Equals("var")) {
            index++;
            var declareVar = new DeclareVar();

            if (index < tokens.Count &&
                tokens[index] is string) {
                declareVar.Ident = (string)tokens[index];
            }
            else {
                throw new Exception("expected variable name after 'var'");
            }

            index++;

            if (index == tokens.Count ||
                tokens[index] != Scanner.Equal) {
                throw new Exception("expected = after 'var ident'");
            }

            index++;

            declareVar.Expr = ParseExpr();
            result = declareVar;
        }
        else if (tokens[index].Equals("read_int")) {
            index++;
            var readInt = new ReadInt();

            if (index < tokens.Count &&
                tokens[index] is string) {
                readInt.Ident = (string)tokens[index++];
                result = readInt;
            }
            else {
                throw new Exception("expected variable name after 'read_int'");
            }
        }
        else if (tokens[index].Equals("for")) {
            index++;
            var forLoop = new ForLoop();

            if (index < tokens.Count &&
                tokens[index] is string) {
                forLoop.Ident = (string)tokens[index];
            }
            else {
                throw new Exception("expected identifier after 'for'");
            }

            index++;

            if (index == tokens.Count ||
                tokens[index] != Scanner.Equal) {
                throw new Exception("for missing '='");
            }

            index++;

            forLoop.From = ParseExpr();

            if (index == tokens.Count ||
                !tokens[index].Equals("to")) {
                throw new Exception("expected 'to' after for");
            }

            index++;

            forLoop.To = ParseExpr();

            if (index == tokens.Count ||
                !tokens[index].Equals("do")) {
                throw new Exception("expected 'do' after from expression in for loop");
            }

            index++;

            forLoop.Body = ParseStmt();
            result = forLoop;

            if (index == tokens.Count ||
                !tokens[index].Equals("end")) {
                throw new Exception("unterminated 'for' loop body");
            }

            index++;
        }
        else if (tokens[index].Equals("foreach")) {
            index++;
            var foreach_ = new Foreach();

            if (index < tokens.Count &&
                tokens[index] is string) {
                foreach_.Ident = (string)tokens[index];
            }
            else {
                throw new Exception("expected identifier after 'for'");
            }

            index++;

            if (index == tokens.Count ||
                tokens[index] != Scanner.ForeachIn) {
                throw new Exception("foreach missing 'in'");
            }

            index++;
            if (index < tokens.Count) {
                foreach_.IEnumerable = ParseExpr();
            }
            else {
                throw new Exception("expected IEnumerable after 'for'");
            }

            if (index == tokens.Count ||
                !tokens[index].Equals("do")) {
                throw new Exception("expected 'do' after from expression in for loop");
            }

            index++;

            foreach_.Body = ParseStmt();
            result = foreach_;

            if (index == tokens.Count ||
                !tokens[index].Equals("end")) {
                throw new Exception("unterminated 'foreach' loop body");
            }

            index++;
        }
        else if (tokens[index].Equals("if")) {
            index++;
            var if_ = new If();

            if (index < tokens.Count) {
                if_.Condition = ParseExpr();
            }
            else {
                throw new Exception("expected Condition after 'if'");
            }

            if (index == tokens.Count ||
                !tokens[index].Equals("then")) {
                throw new Exception("expected 'then' after 'if'");
            }

            index++;

            if_.Body = ParseStmt();
            result = if_;

            if (index == tokens.Count ||
                !tokens[index].Equals("end")) {
                throw new Exception("unterminated 'foreach' loop body");
            }

            index++;
        }
        else if (tokens[index] is string) {
            // assignment

            var assign = new Assign();
            assign.Ident = (string)tokens[index++];

            if (index == tokens.Count ||
                tokens[index] != Scanner.Equal) {
                throw new Exception("expected '='");
            }

            index++;

            assign.Expr = ParseExpr();
            result = assign;
        }
        else {
            throw new Exception("parse error at token " + index + ": " + tokens[index]);
        }

        if (index < tokens.Count && tokens[index] == Scanner.Semi) {
            index++;

            if (index < tokens.Count &&
                !tokens[index].Equals("end")) {
                var sequence = new Sequence();
                sequence.First = result;
                sequence.Second = ParseStmt();
                result = sequence;
            }
        }

        return result;
    }

    private Expr ParseExpr() {
        if (index == tokens.Count) {
            throw new Exception("expected expression, got EOF");
        }
        Expr ret = null;
        if (tokens[index] is StringBuilder) {
            var stringLiteral = new StringLiteral();
            stringLiteral.Value = (tokens[index++]).ToString()
                .Replace(@"\r\n","\r\n");
            ret = stringLiteral;
        }
        else if (tokens[index] is int) {
            var intValue = (int)tokens[index++];
            var intLiteral = new IntLiteral();
            intLiteral.Value = intValue;
            ret = intLiteral;
        }
        else if (tokens[index] is string) {
            string value = (string)tokens[index];
            if (value.Equals("match")) {
                index++;
                var match = new Match();

                if (index < tokens.Count) {
                    match.Input = ParseExpr();
                }
                else {
                    throw new Exception("expected input after 'match'");
                }
                if (index < tokens.Count) {
                    match.Pattern = ParseExpr();
                    ret = match;
                }
                else
                    throw new Exception("expected pattern after 'match'");
            }
            else if (value.Equals("newsb")) {
                var builder = new Builder();
                index++;
                ret = builder;
            }
            else if (value.Equals("len")) {
                var len = new StrLen();
                index++;
                if (index < tokens.Count) {
                    len.Input = ParseExpr();
                }
                else {
                    throw new Exception("expected input after 'len'");
                }
                ret = len;
            }
            else {
                var ident = (string)tokens[index++];
                Variable vari = new Variable();
                vari.Ident = ident;
                ret = vari;
            }
            if ((index) < tokens.Count && tokens[index] == Scanner.Eq) {
                BinExpr binexp = new BinExpr();
                binexp.Left = ret;
                binexp.Op = BinOp.Eq;
                index++;
                if (index < tokens.Count) {
                    binexp.Right = ParseExpr();
                }
                else {
                    throw new Exception("expected right after '=='");
                }
                ret = binexp;
            }
            return ret;
        }
        else {
            throw new Exception("expected string literal, int literal, or variable");
        }
        return ret;
    }
}